/*
 * singleton.hpp
 * Copyright (c) 2011 - TÖRÖK Attila (torokati44@gmail.com)
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 *   1. The origin of this software must not be misrepresented; you must not
 *      claim that you wrote the original software. If you use this software
 *      in a product, an acknowledgment in the product documentation would be
 *      appreciated but is not required.
 *
 *   2. Altered source versions must be plainly marked as such, and must not be
 *      misrepresented as being the original software.
 *
 *   3. This notice may not be removed or altered from any source
 *      distribution.
 */

// This file is based on OGRE's OgreSingleton.h which can be found here:
// http://ogre.svn.sourceforge.net/viewvc/ogre/branches/v1-7/OgreMain/include/OgreSingleton.h?revision=9527&view=markup
// Thanks!

/*
This is a basic easy-to use template base class for the singleton design pattern.

Usage:

class MySingleton : public Singleton< MySingleton > // <- important!
{
	// your members, functions, etc
	// DESTRUCTOR MUST BE VIRTUAL!
};

You have to instantiate one (and only one) object of MySingleton class.
Preferably with the new operator, because that will create it in the heap,
and with normal declaration it might get out of scope and get destructed.
Also, you will have to delete it when you are done.
Declaring in the main() and in the global scope are OK, as long as MySingleton
in the stack is OK.
These singletons can be instantiated and deleted any number of times, but at most
one instance can exist at any time.
*/

#ifndef SINGLETON_HPP_INCLUDED
#define SINGLETON_HPP_INCLUDED

template< class T >
class Singleton
{
private:
	// These two are intentionally undefined, because they are forbidden on a singleton.
	// Maybe I could define them and assert a constant there, but this is just fine, too.
	Singleton(const Singleton< T >&);
	Singleton& operator = (const Singleton< T >&);
	
	// msp means my static pointer ...
	static T* msp_instance;
	
protected:
	Singleton();
	virtual ~Singleton();
	
public:
	inline static T& instance();
	inline static T* instance_ptr();
	
	inline T& operator () ();
};

//template functions have to be defined in every compilation unit...
#include "singleton.inl"

#endif //SINGLETON_HPP_INCLUDED
