/*
 * string.cpp - XD-Ball
 * Copyright (c) 2011, 2012 - TÖRÖK Attila (torokati44@gmail.com)
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 *	 1. The origin of this software must not be misrepresented; you must not
 *			claim that you wrote the original software. If you use this software
 *			in a product, an acknowledgment in the product documentation would be
 *			appreciated but is not required.
 *
 *	 2. Altered source versions must be plainly marked as such, and must not be
 *			misrepresented as being the original software.
 *
 *	 3. This notice may not be removed or altered from any source
 *			distribution.
 */

//MODIFIED FROM KUKATZ 3D

#include "string.hpp"

#include <iostream>

//#include <SFML/Graphics.hpp>

#include "resources.hpp"
#include "font.hpp"

String::String(const sf::Unicode::Text& t, const std::string& f, float s):
	font_id(f), text(t), contains_color_reset(false), color(255, 255, 255, 255),
	font_ptr(Resources::instance().fonts[f]), billboard(false), size(s)
{
	displaylist = glGenLists(1);
	make_list();
}

String::String(const sf::Unicode::Text& t, const std::string& f):
	font_id(f), text(t), contains_color_reset(false), color(255, 255, 255, 255),
	font_ptr(Resources::instance().fonts[f]), billboard(false),
	size(font_ptr->line_height)
{
	displaylist = glGenLists(1);
	make_list();
}

void String::make_list()
{
	contains_color_reset = false;
	
	sf::Vector2i cursor(0, font_ptr->size);
	
	Font::glyph_t* prev_glyph_ptr = 0;
	Font::glyph_t* glyph_ptr = 0;
	
	glNewList(displaylist, GL_COMPILE);
	glBegin(GL_QUADS);
	for (size_t i = 0; i < text.length(); ++i)
	{
		prev_glyph_ptr = glyph_ptr;
		glyph_ptr = &(font_ptr->glyphs[text[i]]);
		
		switch (text[i])
		{
			case '\n':
			{
				cursor.x = 0;
				cursor.y -= font_ptr->line_height;
			}
			break;
			case '\t':
			{
				cursor.x /= font_ptr->line_height;
				cursor.x += 1;
				cursor.x *= font_ptr->line_height;
			}
			break;
			case 0xC:
			{
				sf::Uint8 r = text[++i];
				sf::Uint8 g = text[++i];
				sf::Uint8 b = text[++i];
				sf::Uint8 a = text[++i];
				
				glColor4ub(r, g, b, a);
			}
			break;
			case 0xE:
			{
				contains_color_reset = true;
				glColor4ub(color.r, color.g, color.b, color.a);
			}
			break;
			default:
			{
				if (i > 0) // kerning
				{
					for (sf::Uint32 j = 0; j < prev_glyph_ptr->kernings.size(); ++j)
					{
						if (prev_glyph_ptr->kernings[j].first == i)
						{
							cursor.x -= prev_glyph_ptr->kernings[j].second;
							//std::cout << "kern!\n";
							break;
						}
					}
				}
				
				sf::Vector2i fontmap_size(font_ptr->texture_ptr->get_width(),
					font_ptr->texture_ptr->get_height());
				
				int vx;
				int vy;
				
				// BOTTOM LEFT
				glTexCoord2f((float)glyph_ptr->x / (float)fontmap_size.x,
					(float)(glyph_ptr->y + glyph_ptr->h) / (float)fontmap_size.y);
				
				vx = cursor.x + glyph_ptr->xo;
				vy = cursor.y - glyph_ptr->h - glyph_ptr->yo;
				
				glVertex2i(vx, vy);
				
				if (i == 0)
				{
					bottom_left.x = vx;
					bottom_left.y = vy;
				}
				else
				{
					if (vx < bottom_left.x)
					{
						bottom_left.x = vx;
					}
					if (vy < bottom_left.y)
					{
						bottom_left.y = vy;
					}
				}
				
				// BOTTOM RIGHT
				glTexCoord2f((float)(glyph_ptr->x + glyph_ptr->w) / (float)fontmap_size.x,
					(float)(glyph_ptr->y + glyph_ptr->h) / (float)fontmap_size.y);
				glVertex2i(cursor.x + glyph_ptr->w + glyph_ptr->xo, cursor.y - glyph_ptr->h - glyph_ptr->yo);
				
				// TOP RIGHT
				glTexCoord2f((float)(glyph_ptr->x + glyph_ptr->w) / (float)fontmap_size.x,
					(float)(glyph_ptr->y) / (float)fontmap_size.y);
				
				vx = cursor.x + glyph_ptr->w + glyph_ptr->xo;
				vy = cursor.y - glyph_ptr->yo;
				
				glVertex2i(vx, vy);
				
				if (i == 0)
				{
					top_right.x = vx;
					top_right.y = vy;
				}
				else
				{
					if (vx > top_right.x)
					{
						top_right.x = vx;
					}
					if (vy > top_right.y)
					{
						top_right.y = vy;
					}
				}
				
				// TOP LEFT
				glTexCoord2f((float)glyph_ptr->x / (float)fontmap_size.x,
					(float)(glyph_ptr->y) / (float)fontmap_size.y);
				glVertex2i(cursor.x + glyph_ptr->xo, cursor.y - glyph_ptr->yo);
				
				cursor.x += glyph_ptr->xa;
			}
			break;
		}
	}
	glEnd();
	glEndList();
}

float String::get_width()
{
	return (float)(top_right.x - bottom_left.x) * (float)size / (float)font_ptr->line_height;
}

float String::get_height()
{
	return (float)(top_right.y - bottom_left.y) * (float)size / (float)font_ptr->line_height;
}

unsigned int String::num_lines()
{
	unsigned int ret = 1;
	for (size_t i = 0; i < text.size(); ++i)
	{
		if (text[i] == '\n')
		{
			++ret;
		}
	}
	return ret;
}

void String::render()
{
	font_ptr->texture_ptr->bind();
	
	glPushMatrix();
	float scale = (float)size / (float)(font_ptr->line_height);
	
	if (billboard)
	{
		float modelview[16];
		
		glGetFloatv(GL_MODELVIEW_MATRIX, modelview);
		
		for (unsigned int i = 0; i < 3; i++)
		{
			for ( unsigned int j = 0; j < 3; j++)
			{
				modelview[i * 4 + j] = ((i == j)?(scale / 200.f):0.0f);
			}
		}
		
		glLoadMatrixf(modelview);
		glTranslatef(-get_width() / 2.0f, -get_height() / 2.0f, 0);
	}
	else
	{
		glScalef(scale, scale, scale);
	}
	glTranslatef(0, (num_lines() - 1) * font_ptr->line_height, 0);
	glColor4ub(color.r, color.g, color.b, color.a);
	
	glCallList(displaylist);
	glColor4ub(255, 255, 255, 255);
	
	glPopMatrix();
}

void String::set_text(const sf::Unicode::Text& t)
{
	if (text.compare(t) != 0)
	{
		text = t;
		make_list();
	}
}

sf::Unicode::Text String::get_text()
{
	return text;
}

void String::set_color(const sf::Color& c)
{
	color = c;
	
	if (contains_color_reset)
	{
		make_list();
	}
}

void String::set_color(sf::Uint8 r, sf::Uint8 g, sf::Uint8 b, sf::Uint8 a)
{
	color.r = r;
	color.g = g;
	color.b = b;
	color.a = a;
	
	if (contains_color_reset)
	{
		make_list();
	}
}

const sf::Color& String::get_color()
{
	return color;
}

String::~String()
{
	glDeleteLists(displaylist, 1);
}
